@@ -2,7 +2,7 @@
    +----------------------------------------------------------------------+
    | Xdebug                                                               |
    +----------------------------------------------------------------------+
-   | Copyright (c) 2002-2019 Derick Rethans                               |
+   | Copyright (c) 2002-2020 Derick Rethans                               |
    +----------------------------------------------------------------------+
    | This source file is subject to version 1.01 of the Xdebug license,   |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -289,7 +289,7 @@ void xdebug_debugger_throw_exception_hook(zend_class_entry * exception_ce, zval
 	/* Start JIT if requested and not yet enabled */
 	xdebug_do_jit();
 
-	if (xdebug_is_debug_connection_active_for_current_pid()) {
+	if (xdebug_is_debug_connection_active_for_current_pid() && XG_DBG(breakpoints_allowed)) {
 		int exception_breakpoint_found = 0;
 
 		/* Check if we have a wild card exception breakpoint */
@@ -504,9 +504,9 @@ void xdebug_debugger_rinit(void)
 		XG_DBG(ide_key) = xdstrdup(idekey);
 	}
 
-	XG_DBG(no_exec)        = 0;
-	XG_LIB(active_symbol_table) = NULL;
-	XG_LIB(This) = NULL;
+	XG_DBG(no_exec) = 0;
+	xdebug_lib_set_active_symbol_table(NULL);
+	xdebug_lib_set_active_object(NULL);
 
 	/* Check if we have this special get variable that stops a debugging
 	 * request without executing any code */
@@ -531,7 +531,10 @@ void xdebug_debugger_rinit(void)
 	xdebug_mark_debug_connection_not_active();
 
 	XG_DBG(breakpoints_allowed) = 1;
+	XG_DBG(detached) = 0;
 	XG_DBG(breakable_lines_map) = xdebug_hash_alloc(2048, (xdebug_hash_dtor_t) xdebug_line_list_dtor);
+	XG_DBG(function_count) = 0;
+	XG_DBG(class_count) = 0;
 	XG_DBG(remote_log_file) = NULL;
 
 	/* Initialize some debugger context properties */
@@ -640,18 +643,25 @@ static void add_function_to_lines_list(xdebug_lines_list *lines_list, zend_op_ar
 
 static void resolve_breakpoints_for_function(xdebug_lines_list *lines_list, zend_op_array *opa)
 {
-	if (!ZEND_USER_CODE(opa->type)) {
-		return;
-	}
-
 	add_function_to_lines_list(lines_list, opa);
 }
 
-static void resolve_breakpoints_for_class(xdebug_lines_list *file_function_lines_list, zend_class_entry *ce)
+static void resolve_breakpoints_for_class(xdebug_lines_list *file_function_lines_list, zend_class_entry *ce, zend_string *filename)
 {
 	zend_op_array    *function_op_array;
 
 	ZEND_HASH_FOREACH_PTR(&ce->function_table, function_op_array) {
+		if (!ZEND_USER_CODE(function_op_array->type)) {
+			continue;
+		}
+		/* Only resolve if the file names are the same. This is needed in case
+		 * of inheritance or traits where op arrays from other files might get introduced */
+		if (ZSTR_LEN(filename) != ZSTR_LEN(function_op_array->filename)) {
+			continue;
+		}
+		if (strcmp(ZSTR_VAL(filename), ZSTR_VAL(function_op_array->filename)) != 0) {
+			continue;
+		}
 		resolve_breakpoints_for_function(file_function_lines_list, function_op_array);
 	} ZEND_HASH_FOREACH_END();
 }
@@ -662,7 +672,7 @@ void xdebug_debugger_compile_file(zend_op_array *op_array)
 	zend_class_entry *class_entry;
 	xdebug_lines_list *file_function_lines_list;
 
-	if (!XINI_DBG(remote_enable)) {
+	if (!XINI_DBG(remote_enable) || !XG_DBG(breakable_lines_map)) {
 		return;
 	}
 
@@ -672,6 +682,9 @@ void xdebug_debugger_compile_file(zend_op_array *op_array)
 		if (_idx == XG_DBG(function_count)) {
 			break;
 		}
+		if (!ZEND_USER_CODE(function_op_array->type)) {
+			continue;
+		}
 		resolve_breakpoints_for_function(file_function_lines_list, function_op_array);
 	} ZEND_HASH_FOREACH_END();
 	XG_DBG(function_count) = CG(function_table)->nNumUsed;
@@ -680,7 +693,10 @@ void xdebug_debugger_compile_file(zend_op_array *op_array)
 		if (_idx == XG_DBG(class_count)) {
 			break;
 		}
-		resolve_breakpoints_for_class(file_function_lines_list, class_entry);
+		if (class_entry->type == ZEND_INTERNAL_CLASS) {
+			continue;
+		}
+		resolve_breakpoints_for_class(file_function_lines_list, class_entry, op_array->filename);
 	} ZEND_HASH_FOREACH_END();
 	XG_DBG(class_count) = CG(class_table)->nNumUsed;
 
@@ -701,10 +717,10 @@ static void resolve_breakpoints_for_eval(int eval_id, zend_op_array *opa)
 	xdebug_lines_list *lines_list;
 	char *eval_filename = xdebug_sprintf("dbgp://%d", eval_id);
 	zend_string *eval_string = zend_string_init(eval_filename, strlen(eval_filename), 0);
-	
+
 	lines_list = get_file_function_line_list(eval_string);
 	add_function_to_lines_list(lines_list, opa);
-		
+
 	resolve_breakpoints_for_function(lines_list, opa);
 
 	if (!xdebug_is_debug_connection_active_for_current_pid()) {
@@ -717,7 +733,7 @@ static void resolve_breakpoints_for_eval(int eval_id, zend_op_array *opa)
 		&(XG_DBG(context)),
 		eval_string
 	);
-		
+
 	zend_string_release(eval_string);
 	xdfree(eval_filename);
 }
